"""
Wigner (rotation) matrices for complex and real spherical harmonics
Evaluation and symbolic convention follows:
 Sakurai, J. J. "Modern Quantum Mechanics 2Nd Edition." Person New International edition (2014).
 Page 238, Eq. 3.9.33.
"""

import math
import os

import torch
from joblib import Memory
from scipy.special import factorial

from torch_gauge import ROOT_DIR

memory = Memory(os.path.join(ROOT_DIR, ".o3_cache"), verbose=0)


@memory.cache
def _wigner_small_d_coeff(j, m1, m, k):
    sgn = (-1) ** (k - m + m1)
    numer = math.sqrt(
        factorial(j + m) * factorial(j - m) * factorial(j + m1) * factorial(j - m1)
    )
    denom = (
        factorial(j + m - k)
        * factorial(k)
        * factorial(j - k - m1)
        * factorial(k - m + m1)
    )
    if denom == 0:
        return None
    else:
        return sgn * numer / denom


def wigner_small_d_csh(j, beta):
    """Rotating complex spherical harmonics by the y axis"""
    small_d = torch.zeros(2 * j + 1, 2 * j + 1, dtype=torch.double)
    for m in range(-j, j + 1):
        for m1 in range(-j, j + 1):
            kmin = min(0, m - m1)
            kmax = max(0, j + m, j - m1)
            d_m1_m = 0
            for k in range(kmin, kmax + 1):
                prefactor = _wigner_small_d_coeff(j, m1, m, k)
                if prefactor is not None:
                    angular = (math.cos(beta / 2)) ** (2 * j - 2 * k + m - m1) * (
                        math.sin(beta / 2)
                    ) ** (2 * k - m + m1)
                    d_m1_m += prefactor * angular
            small_d[m1 + j, m + j] = d_m1_m
    return small_d


def wigner_D_csh(j, alpha, beta, gamma):
    """
    Euler rotation for complex spherical harmonics
    Contra-variant convention, for right multiplication
    """
    ms = torch.arange(-j, j + 1)
    small_d = wigner_small_d_csh(j, beta).to(torch.cdouble)
    alpha_phases = torch.exp(-1j * ms * alpha)
    gamma_phases = torch.exp(-1j * ms * gamma)
    wigner_D = alpha_phases.unsqueeze(1) * small_d * gamma_phases.unsqueeze(0)
    assert torch.allclose(
        wigner_D.mm(wigner_D.t().conj()),
        torch.eye(2 * j + 1, 2 * j + 1, dtype=torch.cdouble),
        atol=1e-7,
    )
    return wigner_D


@memory.cache
def csh_to_rsh(j):
    transform_mat = torch.zeros(2 * j + 1, 2 * j + 1, dtype=torch.cdouble)
    for m in range(-j, j + 1):
        # Enumerate by out m
        if m < 0:
            transform_mat[j + m, j + m] = 1j / math.sqrt(2)
            transform_mat[j - m, j + m] = -((-1) ** m) * 1j / math.sqrt(2)
        elif m == 0:
            transform_mat[j + m, j + m] = 1
        elif m > 0:
            transform_mat[j - m, j + m] = 1 / math.sqrt(2)
            transform_mat[j + m, j + m] = (-1) ** m / math.sqrt(2)
        else:
            raise ValueError
    return transform_mat


def wigner_D_rsh(j, alpha, beta, gamma):
    c2r = csh_to_rsh(j)
    wigner_csh = wigner_D_csh(j, alpha, beta, gamma)
    wigner_rsh = (c2r.conj().t()).mm(wigner_csh.t().conj()).mm(c2r)
    # Checking the RSH rotation matrix entries are all real
    assert torch.allclose(
        wigner_rsh.imag, torch.zeros(2 * j + 1, 2 * j + 1, dtype=torch.double)
    ), print(c2r, wigner_csh, wigner_rsh)
    return wigner_rsh.real



"""
SO(3) Clebsch-Gordan coefficients and the coupler
See Page 225, Eq. 3.8.49 of J. J. Sakurai, and Eq. 6, 9 of
 Schulten, Klaus, and Roy G. Gordon, Journal of Mathematical Physics 16.10 (1975): 1961-1970
"""


import os

import torch
from joblib import Memory
from sympy import N
from sympy.physics.quantum.cg import CG

from torch_gauge import ROOT_DIR
from torch_gauge.o3.spherical import SphericalTensor

memory = Memory(os.path.join(ROOT_DIR, ".o3_cache"), verbose=0)


class LeviCivitaCoupler(torch.nn.Module):
    """
    Simple tensor coupling module when max_l==1.
    The input spherical tensors must have n_rep_dims==1 and aligned dimensions
    """

    def __init__(self, metadata: torch.LongTensor):
        super().__init__()
        assert metadata.dim() == 1
        assert (
            len(metadata) == 2
        ), "Only SphericalTensor of max degree 1 is applicable for Cevi-Levita"
        self._metadata = metadata

    def forward(self, x1: SphericalTensor, x2: SphericalTensor, overlap_out=True):
        assert x1.metadata.shape[0] == 1
        assert x2.metadata.shape[0] == 1
        assert x1.rep_dims[0] == x2.rep_dims[0]
        coupling_dim = x1.rep_dims[0]
        assert torch.all(x1.metadata[0].eq(self._metadata))
        assert torch.all(x2.metadata[0].eq(self._metadata))
        ten_l1_1 = x1.ten.narrow(
            coupling_dim, self._metadata[0], self._metadata[1] * 3
        ).unflatten(coupling_dim, (3, self._metadata[1]))
        ten_l1_2 = x2.ten.narrow(
            coupling_dim, self._metadata[0], self._metadata[1] * 3
        ).unflatten(coupling_dim, (3, self._metadata[1]))
        # 0x0->0
        out_000 = x1.ten.narrow(coupling_dim, 0, self._metadata[0]) * x2.ten.narrow(
            coupling_dim, 0, self._metadata[0]
        )
        # 0x1->1
        out_011 = (
            x1.ten.narrow(coupling_dim, 0, self._metadata[1])
            .unsqueeze(coupling_dim)
            .mul(ten_l1_2)
        )
        # 1x0->1
        out_101 = (
            x2.ten.narrow(coupling_dim, 0, self._metadata[1])
            .unsqueeze(coupling_dim)
            .mul(ten_l1_1)
        )
        # 1x1->0
        out_110 = (ten_l1_1 * ten_l1_2).sum(coupling_dim)
        # 1x1->1, note that cross works since (y,z,x) is a canonical order
        out_111 = torch.cross(ten_l1_1, ten_l1_2, dim=coupling_dim)
        if overlap_out:
            # Align and contract the coupling outputs
            out_l0 = out_000
            out_l0.narrow(coupling_dim, 0, self._metadata[1]).add(out_110)
            out_l1 = (
                out_111.add(out_101)
                .add(out_011)
                .flatten(coupling_dim, coupling_dim + 1)
            )
            return x1.self_like(torch.cat([out_l0, out_l1], dim=coupling_dim))
        else:
            # Concatenate the coupling outputs to form a augmented tensor
            out_l0 = torch.cat([out_000, out_110], dim=coupling_dim)
            out_l1 = torch.cat(
                [out_101, out_011, out_111], dim=coupling_dim + 1
            ).flatten(coupling_dim, coupling_dim + 1)
            return SphericalTensor(
                torch.cat([out_l0, out_l1], dim=coupling_dim),
                rep_dims=(coupling_dim,),
                metadata=torch.LongTensor(
                    [[self._metadata[0] + self._metadata[1], self._metadata[1] * 3]]
                ),
            )


@memory.cache
def get_clebsch_gordan_coefficient(j1, j2, j, m1, m2, m):
    """
    Generate Clebsch-Gordan coefficients using sympy with caching
    """
    return N(CG(j1, m1, j2, m2, j, m).doit())



import torch

from torch_gauge.o3.spherical import SphericalTensor


class SSP(torch.nn.Softplus):
    """Shifted SoftPlus activation"""

    def __init__(self, beta=1, threshold=20):
        super().__init__(beta, threshold)

    def forward(self, input):
        return F.softplus(input, self.beta, self.threshold) - math.sqrt(2)


class Swish_fn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, i):
        result = i * torch.sigmoid(i)
        ctx.save_for_backward(i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        i = ctx.saved_tensors[0]
        sigmoid_i = torch.sigmoid(i)
        return grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i)))


class Swish(torch.nn.Module):
    def forward(self, input_tensor):
        return Swish_fn.apply(input_tensor)


class IELin(torch.nn.Module):
    r"""
    Irrep-wise Equivariant Linear Layer

    This module takes a spherical tensor and perform linear transformation within
     the feature channels spanned by each irreducible representation index, (l, m):

    \f[
        \mathbf{h}^{\textrm{out}}_{l,m} = \mathbf{W}^{l} \cdot \mathbf{h}^{\textrm{in}}_{l,m}
    \f]

    The matrix multiplication is always performed on the last dimension of the spherical tensors.

    To vectorize: the coefficient matrix \f$\mathbf{W}^{l}\f$ is stored as a flattened 1D tensor,
     together with a block-sparse layout tensor in the COO format; the coefficients
     are gathered to the block-sparse matrix and converted to the CSR format in runtime.
     May be only beneficial when the batch size is small.

    In the future, we should make it a CUDA kernel to avoid serial operations or redundant
     type conversions for sparse matrices. The first optimization we should do is to
     enable ILP by executing multiple CUDA streams.
    Attributes:
        metadata_in (torch.LongTensor): the number of non-degenerate feature channels of the input tensor
         for each l (and the associated m(s)).
        metadata_out (torch.LongTensor): the number of non-degenerate feature channels of the output tensor
         for each l (and the associated m(s)).a
    Returns:
        The data tensor of the output spherical tensor.
    """

    def __init__(self, metadata_in, metadata_out):
        super().__init__()
        assert metadata_in.dim() == 1
        assert len(metadata_in) == len(metadata_out)
        self._metadata_in = metadata_in
        self._metadata_out = metadata_out
        n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0)) * 2 + 1
        # The bias must be False
        self.linears = torch.nn.ModuleList(
            [
                torch.nn.Linear(
                    metadata_in[l] * n_irreps_per_l[l],
                    metadata_out[l] * n_irreps_per_l[l],
                    bias=False,
                )
                if self._metadata_out[l] > 0
                else None
                for l, _ in enumerate(metadata_in)
            ]
        )
        self._end_inds_in = torch.cumsum(self._metadata_in * n_irreps_per_l, dim=0)
        self._start_inds_in = torch.cat([torch.LongTensor([0]), self._end_inds_in[:-1]])
        self.out_layout = SphericalTensor.generate_rep_layout_1d_(self._metadata_out)
        self.num_out_channels = torch.sum(self._metadata_out).item()

    def forward(self, x: SphericalTensor) -> SphericalTensor:
        assert x.rep_dims[-1] == x.ten.dim() - 1
        assert torch.all(x.metadata[-1].eq(self._metadata_in)), (
            f"Expected the SphericalTensor x and self._metadata_in to have the "
            f"same irrep metadata along the last dimension, got {x.metadata[-1]}"
            f" and {self._metadata_in} instead"
        )
        # TODO: vectorization
        outs = []
        for l, linear_l in enumerate(self.linears):
            if linear_l is None:
                continue
            in_l = x.ten[
                ..., self._start_inds_in[l] : self._end_inds_in[l]
            ].contiguous()
            out_l = linear_l(in_l)
            outs.append(out_l)

        out_ten = torch.cat(outs, dim=-1)
        out_metadata = x.metadata.clone()
        out_metadata[-1] = self._metadata_out
        out_rep_layout = x.rep_layout[:-1] + (self.out_layout,)
        return SphericalTensor(
            out_ten,
            rep_dims=x.rep_dims,
            metadata=out_metadata,
            rep_layout=out_rep_layout,
            num_channels=x.num_channels[:-1] + (self.num_out_channels,),
        )


class RepNorm1d(torch.nn.Module):
    r"""
    The (experimental) Representation Normalization layer.

    \f[
        RepNorm(\mathbf{h})_{l,m} = \frac{BatchNorm(||\mathbf{h}_l||)}{\sqrt{2l+1}} \cdot
         \frac{\mathbf{h}_{l}}{ |(1-\bm{\beta})||\mathbf{h}_{l}|| + \bm{\beta}| + \epsilon}
    \f]

    Heuristically, the trainable \f$\beta\f$ controls the fraction of norm information to be retained.

    Returns:
        normed_ten: the normed invariant content
        gauge_ten: the "pure gauge" spherical tensor
    """

    def __init__(self, num_channels, momentum=0.1, eps=1e-2):
        super().__init__()
        self._num_channels = num_channels
        self._eps = eps
        self.batchnorm = torch.nn.BatchNorm1d(
            num_features=num_channels, momentum=momentum
        )
        # TODO: initialization schemes
        self.beta = torch.nn.Parameter(torch.rand(self._num_channels))

    def forward(self, x: SphericalTensor) -> (torch.Tensor, SphericalTensor):
        x0 = x.invariant()
        assert x0.dim() == 2
        x1 = self.batchnorm(x0)
        divisor = torch.abs(x0.mul(1 - self.beta) + self.beta) + self._eps
        x2 = x.scalar_mul(1 / divisor)
        return x1, x2



import os
from torch_gauge import o3, verlet_list
from torch_gauge._version import get_versions

__version__ = get_versions()["version"]
del get_versions

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))



from setuptools import setup, find_packages
import versioneer

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="torch_gauge",
    description="A light-weight PyTorch extension for gauge-equivariant geometric learning",
    author="Zhuoran Qiao",
    author_email="zqiao@caltech.edu",
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=['torch>=1.7.0'],
)



import torch


class NormContraction1d(torch.autograd.Function):
    """
    Calculate the (stabilized) channel-wise L2-norm of a 1d spherical tensor
    The representation dimension must be an integer
    The gradients at zero are enforced to be 0 by a non-negative eps
    """

    @staticmethod
    def forward(
        ctx, data_ten: torch.Tensor, idx_ten: torch.LongTensor, out_shape, dim, eps
    ):
        sum_sqr = torch.zeros(
            out_shape, device=data_ten.device, dtype=data_ten.dtype
        ).scatter_add_(dim=dim, index=idx_ten, src=data_ten.pow(2))
        norm_shifted = (sum_sqr + eps ** 2).sqrt()
        ctx.dim = dim
        ctx.save_for_backward(data_ten, idx_ten, norm_shifted)
        return norm_shifted - eps

    @staticmethod
    def backward(ctx, grad_output):
        data_ten, dst_ten, norm_shifted = ctx.saved_tensors
        gathered_grad_output = torch.gather(grad_output, ctx.dim, dst_ten)
        gathered_norm_shifted = torch.gather(norm_shifted, ctx.dim, dst_ten)
        norm_grad = data_ten / gathered_norm_shifted
        grad_input = gathered_grad_output * norm_grad
        return grad_input, None, None, None, None


class NormContraction2d(torch.autograd.Function):
    """
    Calculate the (stabilized) channel-wise L2 matrix-norm of a 2d spherical tensor
    The representation dimensions must be a 2-tuple (i, i+1)
    """

    @staticmethod
    def forward(
        ctx,
        data_ten: torch.Tensor,
        idx_tens: torch.LongTensor,
        out_shape,
        dims,
        eps,
    ):
        shape_rep_out = tuple(out_shape[d] for d in dims)
        cache_inds = (
            idx_tens[0] * shape_rep_out[1] + idx_tens[1]
        )  # flattened dst indices
        cache_inds = cache_inds.flatten(*dims)
        sum_sqr = torch.zeros(
            out_shape,
            device=data_ten.device,
            dtype=data_ten.dtype,
        ).flatten(*dims)
        sum_sqr = sum_sqr.scatter_add_(
            dim=dims[0], index=cache_inds, src=data_ten.flatten(*dims).pow(2)
        )
        norm_cache_shifted = (sum_sqr + eps ** 2).sqrt()
        norm_shifted = norm_cache_shifted.view(
            out_shape
        )  # must be contiguous at this point
        ctx.dims = dims
        ctx.save_for_backward(data_ten, cache_inds, norm_cache_shifted)
        return norm_shifted - eps

    @staticmethod
    def backward(ctx, grad_output):
        data_ten, cache_inds, norm_cache = ctx.saved_tensors
        dims = ctx.dims
        gathered_grad_output = torch.gather(
            grad_output.flatten(*dims),
            dims[0],
            cache_inds,
        )
        gathered_norm_shifted = torch.gather(norm_cache, dims[0], cache_inds)
        gathered_norm_shifted = gathered_norm_shifted.view(data_ten.shape)
        norm_grad = data_ten / gathered_norm_shifted
        grad_input = gathered_grad_output.view_as(norm_grad) * norm_grad
        return grad_input, None, None, None, None


class IrrepBroadcast(torch.autograd.Function):
    """
    Specialized functional for broadcasting scalar features with
     a SO(3) kernel of identical channel-sizes per angular quantum number l
    """

    @staticmethod
    def forward(ctx, kernel, feature, metadata):
        raise NotImplementedError



"""
Testing real spherical harmonics generation
"""

import math
import random

import torch

from torch_gauge.geometric import rotation_matrix_xyz
from torch_gauge.o3.rsh import RSHxyz
from torch_gauge.o3.wigner import wigner_D_rsh, wigner_small_d_csh


def test_batch_rsh_generation():
    rshmodule = RSHxyz(max_l=6)
    xyz = torch.normal(0, 1, size=(1024, 3))
    rshs = rshmodule(xyz)
    assert torch.all(rshs.metadata.eq(torch.tensor([[1, 1, 1, 1, 1, 1, 1]])))


def test_rsh_explicit():
    rshmodule = RSHxyz(max_l=2)
    # Compare with the explicit formula up to l=2
    for _ in range(100):
        x = torch.normal(0, 1, size=(1,)).double()
        y = torch.normal(0, 1, size=(1,)).double()
        z = torch.normal(0, 1, size=(1,)).double()
        r = torch.sqrt(x ** 2 + y ** 2 + z ** 2)
        rshs_ref = torch.tensor(
            [
                1,
                y,
                z,
                x,
                math.sqrt(3) * x * y,
                math.sqrt(3) * y * z,
                1 / 2 * (3 * z ** 2 - r ** 2),
                math.sqrt(3) * x * z,
                (math.sqrt(3) / 2) * (x ** 2 - y ** 2),
            ]
        )
        assert torch.allclose(
            rshmodule(torch.stack([x, y, z], dim=1)).ten,
            rshs_ref.unsqueeze(0),
            atol=1e-8,
            rtol=0,
        )


def test_rsh_batch_explicit():
    rshmodule = RSHxyz(max_l=2)
    x = torch.normal(0, 1, size=(1024,)).double()
    y = torch.normal(0, 1, size=(1024,)).double()
    z = torch.normal(0, 1, size=(1024,)).double()
    r = torch.sqrt(x ** 2 + y ** 2 + z ** 2)
    rshs_ref = torch.stack(
        [
            torch.ones_like(x),
            y,
            z,
            x,
            math.sqrt(3) * x * y,
            math.sqrt(3) * y * z,
            1 / 2 * (3 * z ** 2 - r ** 2),
            math.sqrt(3) * x * z,
            (math.sqrt(3) / 2) * (x ** 2 - y ** 2),
        ],
        dim=1,
    )
    assert torch.allclose(
        rshmodule(torch.stack([x, y, z], dim=1)).ten,
        rshs_ref,
        atol=1e-8,
        rtol=0,
    )


def test_wigner_small_d():
    # Check against table https://www.wikiwand.com/en/Wigner_D-matrix#/List_of_d-matrix_elements
    theta = random.random() * 2 * math.pi
    ref_mat = torch.tensor(
        [
            [
                1 / 2 * (1 + math.cos(theta)),
                1 / math.sqrt(2) * math.sin(theta),
                1 / 2 * (1 - math.cos(theta)),
            ],
            [
                -1 / math.sqrt(2) * math.sin(theta),
                math.cos(theta),
                1 / math.sqrt(2) * math.sin(theta),
            ],
            [
                1 / 2 * (1 - math.cos(theta)),
                -1 / math.sqrt(2) * math.sin(theta),
                1 / 2 * (1 + math.cos(theta)),
            ],
        ],
        dtype=torch.double,
    )
    assert torch.allclose(
        wigner_small_d_csh(1, theta),
        ref_mat,
        atol=1e-7,
    )


def test_wigner_rsh_z():
    # Along z axis
    for _ in range(10):
        alpha = random.random() * 2 * math.pi
        rsh_wigner_1 = wigner_D_rsh(1, alpha, 0, 0)
        # In the y-z-x order
        ref_rotmat = torch.tensor(
            [
                [math.cos(alpha), 0, -math.sin(alpha)],
                [0, 1, 0],
                [math.sin(alpha), 0, math.cos(alpha)],
            ],
            dtype=torch.double,
        )
        assert torch.allclose(rsh_wigner_1, ref_rotmat, atol=1e-6, rtol=1e-5,), print(
            "Wigner-D at j=1 does not match the Cartesian rotation matrix: ",
            rsh_wigner_1,
            ref_rotmat,
        )


def test_wigner_rsh_y():
    # Along y axis
    for _ in range(10):
        beta = random.random() * 2 * math.pi
        rsh_wigner_1 = wigner_D_rsh(1, 0, beta, 0)
        # In the y-z-x order
        ref_rotmat = torch.tensor(
            [
                [1, 0, 0],
                [0, math.cos(beta), math.sin(beta)],
                [0, -math.sin(beta), math.cos(beta)],
            ],
            dtype=torch.double,
        )
        assert torch.allclose(rsh_wigner_1, ref_rotmat, atol=1e-6, rtol=1e-5,), print(
            "Wigner-D at j=1 does not match the Cartesian rotation matrix: ",
            rsh_wigner_1,
            ref_rotmat,
        )


def test_wigner_rsh_zyz():
    # Test Euler rotation
    # Note that space-fixed rotation (gamma, beta, alpha) == Euler rotation (alpha, beta, gamma),
    # as elaborated in Sec. 3.3, Page 175-177 of J. J. Sakurai
    for _ in range(10):
        alpha = random.random() * 2 * math.pi
        beta = random.random() * 2 * math.pi
        gamma = random.random() * 2 * math.pi
        rsh_wigner_1 = wigner_D_rsh(1, alpha, beta, gamma)
        ref1 = torch.tensor(
            [
                [math.cos(alpha), 0, -math.sin(alpha)],
                [0, 1, 0],
                [math.sin(alpha), 0, math.cos(alpha)],
            ],
            dtype=torch.double,
        )
        ref2 = torch.tensor(
            [
                [1, 0, 0],
                [0, math.cos(beta), math.sin(beta)],
                [0, -math.sin(beta), math.cos(beta)],
            ],
            dtype=torch.double,
        )
        ref3 = torch.tensor(
            [
                [math.cos(gamma), 0, -math.sin(gamma)],
                [0, 1, 0],
                [math.sin(gamma), 0, math.cos(gamma)],
            ],
            dtype=torch.double,
        )
        ref_rotmat = ref3.mm(ref2).mm(ref1)
        assert torch.allclose(rsh_wigner_1, ref_rotmat, atol=1e-6, rtol=1e-5,), print(
            "Wigner-D at j=1 does not match the Cartesian rotation matrix: ",
            rsh_wigner_1,
            ref_rotmat,
        )


def test_wigner_rsh_rotation():
    # Tight checking up to l=10
    rshmodule = RSHxyz(max_l=10)
    xyz = torch.normal(0, 1, size=(1, 3), dtype=torch.double)
    xyz /= xyz.norm(dim=1).unsqueeze(1)

    alpha = random.random() * 2 * math.pi
    beta = random.random() * 2 * math.pi
    gamma = random.random() * 2 * math.pi
    ref1 = rotation_matrix_xyz(alpha, "z", dtype=torch.double).t()
    ref2 = rotation_matrix_xyz(beta, "y", dtype=torch.double).t()
    ref3 = rotation_matrix_xyz(gamma, "z", dtype=torch.double).t()
    ref_rot = ref3.mm(ref2).mm(ref1)
    rot_xyz = xyz.mm(ref_rot)

    rsh_pre = rshmodule(xyz).ten
    rsh_rot = rshmodule(rot_xyz).ten

    wigner_rot_rshs = []
    for l in range(11):
        real_wigner_l = wigner_D_rsh(l, alpha, beta, gamma)
        wigner_rot_rshs.append(rsh_pre[:, l ** 2 : (l + 1) ** 2].mm(real_wigner_l))

    wigner_rot_rshs = torch.cat(wigner_rot_rshs, dim=1)
    assert torch.allclose(rsh_rot, wigner_rot_rshs, atol=1e-6), print(
        rsh_rot, wigner_rot_rshs
    )



import torch

from torch_gauge.o3.clebsch_gordan import (
    LeviCivitaCoupler,
    get_clebsch_gordan_coefficient,
)
from torch_gauge.o3.spherical import SphericalTensor


def test_generate_cg():
    max_j = 4
    for j1 in range(max_j + 1):
        for j2 in range(max_j + 1):
            for j in range(abs(j1 - j2), max_j + 1):
                print(
                    f"Generating Clebsch-Gordan coefficients for j1={j1}, j2={j2}, j={j}"
                )
                for m1 in range(-j1, j1 + 1):
                    for m2 in range(-j2, j2 + 1):
                        get_clebsch_gordan_coefficient(j1, j2, j, m1, m2, m1 + m2)


def test_levi_civita():
    metadata = torch.LongTensor([[24, 12]])
    dten1 = torch.rand(4, 6, 2, 60, 7)
    dten2 = torch.rand(4, 6, 2, 60, 7)
    spten1 = SphericalTensor(dten1, (3,), metadata)
    spten2 = SphericalTensor(dten2, (3,), metadata)
    coupler = LeviCivitaCoupler(metadata[0])
    overlap_out = coupler(spten1, spten2, overlap_out=True)
    assert overlap_out.ten.shape == (4, 6, 2, 60, 7)
    cat_out = coupler(spten1, spten2, overlap_out=False)
    assert cat_out.ten.shape == (4, 6, 2, 144, 7)
    assert torch.all(cat_out.metadata.eq(torch.LongTensor([[36, 36]])))
